package net.sf.sido.support;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import net.sf.jstring.JStrings;
import net.sf.sido.*;
import net.sf.sido.constraint.ConstraintFactory;
import net.sf.sido.constraint.support.DefaultConstraintFactory;
import net.sf.sido.editor.BooleanEditor;
import net.sf.sido.editor.DataEditor;
import net.sf.sido.editor.DateTimeEditor;
import net.sf.sido.editor.DecimalEditor;
import net.sf.sido.editor.DurationEditor;
import net.sf.sido.editor.PeriodEditor;
import net.sf.sido.editor.SimpleEditor;
import net.sf.sido.editor.StringEditor;
import net.sf.sido.DOCannotFindClassForTypeException;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.joda.time.DateTime;
import org.joda.time.Duration;
import org.joda.time.Period;

import com.google.common.collect.ImmutableList;

public class DefaultDataFactory implements DataFactory {

    static {
        JStrings.add("sido-bundle");
    }

	public static Map<String, Class<?>> defaultAliases() {
		Map<String, Class<?>> aliases = new HashMap<String, Class<?>>();
		aliases.put("boolean", Boolean.class);
		aliases.put("byte", Byte.class);
		aliases.put("short", Short.class);
		aliases.put("integer", Integer.class);
		aliases.put("long", Long.class);
		aliases.put("float", Float.class);
		aliases.put("double", Double.class);
		aliases.put("decimal", BigDecimal.class);
		aliases.put("datetime", DateTime.class);
		aliases.put("period", Period.class);
		aliases.put("duration", Duration.class);
		aliases.put("string", String.class);
		return aliases;
	}

	private Map<String, Class<?>> aliases = defaultAliases();

	private final Map<Class<?>, DataEditor<?>> editors = new HashMap<Class<?>, DataEditor<?>>();
	private ConstraintFactory constraintFactory = new DefaultConstraintFactory();
	private final Map<String, DataSchema> schemas = new LinkedHashMap<String, DataSchema>();

	private final List<DataObjectController> controllers;

	public DefaultDataFactory(DataObjectController... controllers) {
		registerDefaultEditors();
		this.controllers = Arrays.asList(controllers);
	}

	public DefaultDataFactory(List<DataObjectController> controllers) {
		registerDefaultEditors();
		this.controllers = ImmutableList.copyOf(controllers);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see net.sf.sido.io.json.JSONDataFactory#forAlias(java.lang.String)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public <T> Class<T> forAlias(String type) {
		return (Class<T>) this.aliases.get(type);
	}

	public Map<String, Class<?>> getAliases() {
		return this.aliases;
	}

	@Override
	public <T> Class<T> getSimpleType(String type) {
		Class<T> cls = forAlias(type);
		if (cls != null) {
			return cls;
		} else {
			try {
				@SuppressWarnings("unchecked")
				Class<T> c = (Class<T>) Class.forName(type);
				return c;
			} catch (ClassNotFoundException e) {
				throw new DOCannotFindClassForTypeException(e, type);
			}
		}
	}

	@Override
	public DataSchema createSchema(String name) {
		DefaultDataSchema schema = new DefaultDataSchema(this, name);
		this.schemas.put(name, schema);
		return schema;
	}

	@Override
	public List<DataType> getAllTypes() {
		List<DataType> types = new ArrayList<DataType>();
		// For all schemas
		for (DataSchema schema : getSchemas()) {
			// For all types
			for (DataType type : schema.getTypes()) {
				types.add(type);
			}
		}
		return types;
	}

	public ConstraintFactory getConstraintFactory() {
		return this.constraintFactory;
	}

	@Override
	public <T> DataEditor<T> getEditor(Class<T> type) {
		Validate.notNull(type, "Type is required");
		@SuppressWarnings("unchecked")
		DataEditor<T> editor = (DataEditor<T>) this.editors.get(type);
		if (editor != null) {
			return editor;
		} else {
			throw new IllegalArgumentException(String.format("Cannot find any editor for class %s", type.getName()));
		}
	}

	@Override
	public DataSchema getSchema(String name) {
		return this.schemas.get(name);
	}

	@Override
	public Collection<DataSchema> getSchemas() {
		return ImmutableList.copyOf(this.schemas.values());
	}

	@Override
	public List<DataType> getSubTypes(DataType parentType) {
		List<DataType> subtypes = new ArrayList<DataType>();
		// For all types
		for (DataType type : getAllTypes()) {
			if (type.getParentType() == parentType) {
				subtypes.add(type);
			}
		}
		return subtypes;
	}

	@Override
	public DataType getType(String qualifiedName) {
		String schemaName = StringUtils.substringBefore(qualifiedName, DataSchema.QUALIFIER_SEPARATOR);
		String typeName = StringUtils.substringAfter(qualifiedName, DataSchema.QUALIFIER_SEPARATOR);
		// Checks the names
		if (StringUtils.isBlank(schemaName) || StringUtils.isBlank(typeName)) {
			throw new IllegalArgumentException(String.format("Wrong schema%stype qualified name", DataSchema.QUALIFIER_SEPARATOR, qualifiedName));
		}
		// Gets the schema
		DataSchema schema = getSchema(schemaName);
		if (schema != null) {
			return schema.getType(typeName);
		} else {
			return null;
		}
	}

	@Override
	public <T> T onPropertyGet(DataObject o, String name, final T value) {
		T v = value;
		for (DataObjectController controller : this.controllers) {
			v = controller.onPropertyGet(o, name, v);
		}
		return v;
	}

	@Override
	public <T> void onPropertySetPost(DataObject o, String name, final T oldValue, final T newValue) {
		for (DataObjectController controller : this.controllers) {
			controller.onPropertySetPost(o, name, oldValue, newValue);
		}
	}

	@Override
	public <T> T onPropertySetPre(DataObject o, String name, final T oldValue, final T newValue) {
		T v = newValue;
		for (DataObjectController controller : this.controllers) {
			v = controller.onPropertySetPre(o, name, oldValue, v);
		}
		return v;
	}

	protected void registerDefaultEditors() {
		registerEditor(Boolean.class, new BooleanEditor());
		registerEditor(Byte.class, SimpleEditor.get(Byte.class));
		registerEditor(Short.class, SimpleEditor.get(Short.class));
		registerEditor(Integer.class, SimpleEditor.get(Integer.class));
		registerEditor(Long.class, SimpleEditor.get(Long.class));
		registerEditor(Float.class, SimpleEditor.get(Float.class));
		registerEditor(Double.class, SimpleEditor.get(Double.class));
		registerEditor(BigDecimal.class, new DecimalEditor());
		registerEditor(DateTime.class, new DateTimeEditor());
		registerEditor(Period.class, new PeriodEditor());
		registerEditor(Duration.class, new DurationEditor());
		registerEditor(String.class, new StringEditor());
	}

	@Override
	public <T> void registerEditor(Class<T> type, DataEditor<T> editor) {
		this.editors.put(type, editor);
	}

	public void setConstraintFactory(ConstraintFactory constraintFactory) {
		this.constraintFactory = constraintFactory;
	}

	public void setAliases(Map<String, Class<?>> aliases) {
		this.aliases = aliases;
	}

    @Override
    public DataObject newInstance(String type) {
        return getMandatoryType(type).newInstance();
    }

    @Override
    public DataObject newInstance(String type, boolean init) {
        return getMandatoryType(type).newInstance(init);
    }

    @Override
    public DataObject newInstance(String type, boolean init, boolean deep) {
        return getMandatoryType(type).newInstance(init, deep);
    }

    @Override
    public DataType getMandatoryType(String name) {
        DataType type = getType(name);
        if (type != null) {
            return type;
        } else {
            throw new DOTypeNotFoundException(name);
        }
    }
}
